from matplotlib import cm
import numpy as np
import os, sys, glob #, stats , mdtraj_rg
import matplotlib, ast
import matplotlib.pyplot as plt
import re, copyFile
showPlots = True
try:
  os.environ["DISPLAY"] #Detects if display is available
except KeyError:
  showPlots = False
  matplotlib.use('Agg') #Need to set this so doesn't try (and fail) to open interactive graphics window

matplotlib.rc('font', size=7)
matplotlib.rc('axes', titlesize=7)

colors = ['#6495ED','r','#6da81b','#483D8B','#FF8C00','#2E8B57','#800080','#008B8B','#1E90FF']
marker_lib = ['d','^','s','*','p','v']

#-------
#Inputs
#-------

#copy data files to current dir
copy = False
mainDir = '/home/mnguyen/CHE210D/whitmer/PA/'
filetoCopy = ['chargesMC.txt','chargesMD.txt','ionization.txt','PEMC.txt','trajectoryMD.dcd','trajectoryMC.pdb']

#feeding initial values of entropy and Helmholtz free energy
getVal = False
Afiles = ['k0.1_A_final.txt']
Sfiles = ['k0.1_S_final.txt']

#simulation parameters
muPAs = [-8,-6,-4,-2,0,2,4,6,8]
muPBs = []
ks = [0.1]
Ns = [60]
muPAsReweight = np.arange(-10,11,0.5)

#-------------------------------
#copy files to current directory
if copy:
    copyFile.copy(muPAs,muPBs,ks,Ns,filetoCopy,mainDir)

#creating dictionary of data files
#dictionary, key=[k,N,muPA], value = [ionization.txt,PEMC.txt,rg.txt]
fileDict = {}
for N in Ns:
    for k in ks:
        for muPA in muPAs:
                key = (k,N,muPA)
                fHeader = 'k{k}_N{N}_m{muPA}'.format(k=k,N=N, muPA=muPA)
                value = [fHeader + '_ionization.txt', fHeader + '_PEMC.txt', fHeader +'_rg.txt']
                fileDict.update({key:value})

def extractData(n,fId,k,N,muPAs,colId=1):
    """extracting values of an observables over n frames and J simulations from data files
       return a Jxn matrix """
    Xs = []
    nobs = [0]*len(muPAs)
    for j,muPA in enumerate(muPAs):
        key = (k,N,muPA)
        file = fileDict[key][fId]
        f = open(file,'r')
        lines = f.readlines()
        f.close()
        xs = []
        for i,line in enumerate(lines):
            if not line.startswith('#'):
                cols = line.split()
                xs.extend([float(cols[colId])])
        nobs[j] = len(xs)
        Xs.append(xs)
    if not all(x == nobs[0] for x in nobs):
        nobs = np.min(nobs)
        print('number of observations are not the same across simulations, averaging nobs over all simulations')
        print('Using the smallest number of observation')
        Xs = np.array([x[:nobs] for x in Xs],dtype=float)
    else:
        nobs = np.min(nobs)
        Xs = np.array(Xs,dtype=float)
    if n<nobs:
        Xs = np.array([x[:n] for x in Xs],dtype=float)
    return Xs

def histograming(muPAs,k,N,nANbin = None,UNbin = 100,plot=True):
    """Create histogram of potential energy U and number of ionized monomers n_A"""
    obs = []
    nobs = [0]*len(muPAs) #number of observation at each muPA
    Us = []
    nAs = []
    #extracting U and nA from data files
    for j,muPA in enumerate(muPAs):
        key = (k,N,muPA)
        files = fileDict[key]
        f = open(files[0],'r')
        lines1 = f.readlines()
        f = open(files[1],'r')
        lines2 = f.readlines()
        f.close()
        us = []
        nas = []
        for i,line1 in enumerate(lines1):
            line2 = lines2[i] 
            if not line1.startswith('#'):
                cols = line1.split()
                nas.extend([int(N*float(cols[1]))])
                if len(line1)>1:
                    nobs[j] += 1
            if not line2.startswith('#'):
                cols = line2.split()
                us.extend([float(cols[1])])
        nAs.append(nas)
        Us.append(us)
    if not all(x == nobs[0] for x in nobs):
        nobs = np.min(nobs)
        print('number of observations are not the same across simulations, averaging nobs over all simulations')
        print('Using the smallest number of observation')
        Us = np.array([u[:nobs] for u in Us],dtype=float)
        nAs = np.array([na[:nobs] for na in nAs])
    else:
        nobs = np.min(nobs)
        Us = np.array(Us,dtype=float)
        nAs = np.array(nAs)
    print('Number of observations for each simulation: {}'.format(nobs))
    if Us.shape != nAs.shape:
        Exception('mismatch in shapes of U and nA arrays')

    #creating histogram
    dU = (np.max(Us)-np.min(Us))/UNbin
    Ubins = np.array([np.min(Us) + (l+1/2)*dU for l in range(UNbin)])
    if nANbin == None: #each nA value is a bin by default
        nAbins = np.unique(nAs)
    else: 
        dnA = (np.max(nAs)-np.min(nAs))/nANbin
        nAbins = np.array([np.min(nAs) + (l+1/2)*dnA for l in range(nANbin)],dtype=float)
    counts = 0.0001 * np.ones([len(nAbins),len(Ubins)])
    for j in range(len(muPAs)):
        for i in range(int(nobs)):
             nA,U = nAs[j,i],Us[j,i]
             #binning nA
             idx = (np.abs(nAbins - nA)).argmin()
             nA = float(nAbins[idx])
             #binning U
             idx = (np.abs(Ubins - U)).argmin()
             U = Ubins[idx]
             #update obs array with bin values:
             nAs[j,i] = nA
             Us[j,i] = U
             ids = (np.where(abs(nAbins-nA)<1)[0][0],np.where(abs(Ubins-U)<1e-5)[0][0])
             counts[ids] += 1
    if plot:
        plt.figure(figsize=[3,3])
        im=plt.imshow(counts,interpolation = 'spline16',aspect = 'auto',cmap=plt.cm.get_cmap('nipy_spectral'),origin='lower',extent = [min(Ubins),max(Ubins),min(nAbins),max(nAbins)])
        plt.colorbar(im,label='Counts')
        plt.xlabel('$\/PE$')
        plt.ylabel('$\/n_{A-}$')
        plt.savefig('k{}_Counts.png'.format(k),dpi=500,bbox_inches='tight')
        plt.show()
    return nAbins,Ubins,counts,nobs,nAs,Us

def wham(muPAs,kappa,nAbins,Ubins,counts,n,A=[],S=[],tol = 1e-3):
    """Using Weighted Histogram Analysis Method to find the entropy and Helmholtz free energy
        A, S: array with initial values of A and S """
    resA = 10.
    resS = 10.
    resList = []
    muPAs = np.array(muPAs)
    #initial guess for free energies if not providing initial values
    if len(A)==0:
        A = np.zeros(len(muPAs))
    A0 = A.copy()
    #initiate entropy matrix
    if len(S)==0:
        S = np.zeros([len(nAbins),len(Ubins)])
    S0 = S.copy()
    iter = 0
    
    while resA > tol or resS > tol:
        iter += 1
        #update S
        for k,nA in enumerate(nAbins):
            for l,U in enumerate(Ubins): 
                x = -U + muPAs*nA + A
                xmax = np.max(x)
                sum = np.sum(np.exp(x-xmax))
                S[k,l] = np.log(counts[k,l]) - np.log(n) - xmax - np.log(sum)
        #update A
        for j, muPA in enumerate(muPAs):
            x = np.zeros([len(nAbins),len(Ubins)])
            for k,nA in enumerate(nAbins):
                for l,U in enumerate(Ubins):
                    x[k,l] = S[k,l] - U + muPA*nA
            x = x.flatten()
            xmax = np.max(x)
            sum = np.sum(np.exp(x-xmax))
            A[j] = -xmax - np.log(sum)
        A = A - A[0] # set A(j=0) = 0

        resA = np.mean(abs(A-A0))
        resS = np.mean(abs(S-S0))   
        A0 = A.copy()
        S0 = S.copy()
        resList.append([resA,resS])

        if iter%100==0:
            print('Iteration %i residuals of A %6.5f, S %6.5f'%(iter,resA,resS))
        if iter%1000 == 0:
            plt.figure(figsize=[3,3])
            plt.plot(muPAs,A,marker='o',ls='-',lw=0.75,mfc="None",ms=2)
            plt.xlabel('$\mu_{PA}$')
            plt.ylabel('A')
            plt.title('Iteration {}'.format(iter),loc='center')
            plt.savefig('k{}_A.png'.format(kappa),dpi=500,bbox_inches='tight')
 
            plt.figure(figsize=[3,3])
            im=plt.imshow(S,interpolation = 'spline16',aspect = 'auto',cmap=plt.cm.get_cmap('rainbow'),origin='lower',extent = [min(Ubins),max(Ubins),min(nAbins),max(nAbins)])
            plt.colorbar(im,label='$S$')
            plt.xlabel('$\/PE$')
            plt.ylabel('$\/n_{A-}$')
            plt.title('Iteration {}'.format(iter),loc='center')
            plt.savefig('k{}_S.png'.format(kappa),dpi=500,bbox_inches='tight')

            plt.figure(figsize=[3,3])
            plt.plot(range(iter),[a[0] for a in resList ], marker = 'None', ls='-',lw=1, label='A avg residual')
            plt.plot(range(iter),[a[1] for a in resList ], marker = 'None', ls='-',lw=1, label='S avg residual')
            plt.ylim(0,np.min(resList)*10)
            plt.xlabel('Iteration')
            plt.ylabel('Residual')
            plt.legend(loc='best')
            plt.savefig('k{}_Residual.png'.format(kappa),dpi=500,bbox_inches='tight')

            np.savetxt('k{}_S.txt'.format(kappa),S, fmt='%6.7f',header = '# iter=%i resA=%5.5f resS=%5.5f'%(iter,resA,resS))
            np.savetxt('k{}_A.txt'.format(kappa),A, fmt='%6.7f',header = '# iter=%i resA=%5.5f resS=%5.5f'%(iter,resA,resS))

    np.savetxt('k{}_S.txt'.format(kappa),S, fmt='%6.7f',header = '# iter=%i resA=%5.5f resS=%5.5f'%(iter,resA,resS))
    np.savetxt('k{}_A.txt'.format(kappa),A, fmt='%6.7f',header = '# iter=%i resA=%5.5f resS=%5.5f'%(iter,resA,resS))
    print('Finish calculating S and A')
    return S,A

def reweight(muPA,kappa,muPAs,Ubins,nAbins,S,nAs,Us,n,Xs):
    """muPA: muPA to reweight to
       muPAs: muPA values of simulations, length = J
       nAs, Us: J by n  array of nA and U values from J simulations
       n: number of observations made for each muPA simulation
       Xs: matrix of observable X to reweight""" 
    #calculate logarithm of weights
    lnw = np.zeros([len(muPAs),n])
    muPA = float(muPA)
    for j in range(len(muPAs)):
       for i in range(int(n)):
           k,l =  (np.where(abs(nAbins-nAs[j,i])<1)[0][0],np.where(abs(Ubins-Us[j,i])<1e-5)[0][0]) 
           lnw[j,i] = S[k,l] - Us[j,i] + muPA*nAs[j,i]
    lnw = lnw.flatten()
    lnwmax = np.max(lnw)
    Xs = Xs.flatten()
    #get X reweighted at muPA
    X = np.sum(Xs*np.exp(lnw-lnwmax))/np.sum(np.exp(lnw-lnwmax))
    return X 
for i,k in enumerate(ks):    
    nAbins,Ubins,counts,n,nAs,Us = histograming(muPAs,k,N,UNbin = 100)
    Rgs=extractData(n,2,k,N,muPAs)
    if getVal:
        A = []
        S = []
        #get S and A values from file 
        f = open(Afiles[i],'r')
        lines = f.readlines()
        for line in lines:
            if not line.startswith('#') and len(line)>0:
                A.extend([float(a) for a in line.split()])
        f = open(Sfiles[i],'r')
        lines = f.readlines()
        for line in lines:
            if not line.startswith('#') and len(line)>0:
                S.append([float(a) for a in line.split()])
        A = np.array(A)
        S = np.array(S)
        print('initial A \n {}'.format(A))
        print('initial S \n {}'.format(S))
        S,A = wham(muPAs,k,nAbins,Ubins,counts,n,S=S,A=A)
    else:
        S,A = wham(muPAs,k,nAbins,Ubins,counts,n)
    #reweight
    nAReweight = []
    UReweight = []
    RgReweight = []
    for muPA in muPAsReweight: 
        nA = reweight(muPA,k,muPAs,Ubins,nAbins,S,nAs,Us,n,nAs)
        U = reweight(muPA,k,muPAs,Ubins,nAbins,S,nAs,Us,n,Us)
        Rg = reweight(muPA,k,muPAs,Ubins,nAbins,S,nAs,Us,n,Rgs)
        nAReweight.append(nA)
        UReweight.append(U)
        RgReweight.append(Rg)
        obsReweight = [nAReweight,UReweight,RgReweight]
        print('muPA = %i fA = %3.3f'%(muPA,nA/float(60)))
    np.savetxt('k{}_obsReweight.txt'.format(k),np.transpose([muPAsReweight,nAReweight,UReweight,RgReweight]), fmt='%6.7f',header = 'muPA nA U Rg')

    #get simulated data
    UsMean,UsStd = np.mean(Us,axis=1), np.std(Us,axis=1)
    RgsMean,RgsStd = np.mean(Rgs,axis=1), np.std(Rgs,axis=1)
    nAsMean,nAsStd = np.mean(nAs,axis=1), np.std(nAs,axis=1)

    #plot reweighted obs
    plt.figure(figsize=[3,3])
    ax = plt.gca()
    ax.set_prop_cycle('color', colors)
    plt.plot(muPAsReweight,np.array(nAReweight)/float(60),marker=None,ls='-',lw=1,mfc="None",ms=2,label='reweighted')
    plt.errorbar(muPAs,nAsMean/float(60),yerr = nAsStd/float(60),marker='o',ls='-',lw=0.,mfc="None",ms=2,capsize=2,elinewidth=0.5,label='simulated') 
    plt.title('k {}_Reweightedf'.format(k),loc='center')
    plt.legend(loc='best')
    plt.savefig('k{}_Reweightedf.png'.format(k),dpi=500,bbox_inches='tight')

    plt.figure(figsize=[3,3])
    ax = plt.gca()
    ax.set_prop_cycle('color', colors)
    plt.plot(muPAsReweight,np.array(UReweight),marker=None,ls='-',lw=1,mfc="None",ms=2,label='reweighted')
    plt.errorbar(muPAs,UsMean,yerr = UsStd,marker='o',ls='-',lw=0.,mfc="None",ms=2,capsize=2,elinewidth=0.5,label='simulated')
    plt.title('k {}_ReweightedU'.format(k),loc='center')
    plt.legend(loc='best')
    plt.savefig('k{}_ReweightedU.png'.format(k),dpi=500,bbox_inches='tight')

    plt.figure(figsize=[3,3])
    ax = plt.gca()
    ax.set_prop_cycle('color', colors)
    plt.plot(muPAsReweight,np.array(RgReweight),marker=None,ls='-',lw=1,mfc="None",ms=2,label='reweighted')
    plt.errorbar(muPAs,RgsMean,yerr = RgsStd,marker='o',ls='-',lw=0.,mfc="None",ms=2,capsize=2,elinewidth=0.5,label='simulated')
    plt.title('k {}_ReweightedRg'.format(k),loc='center')
    plt.legend(loc='best')
    plt.savefig('k{}_ReweightedRg.png'.format(k),dpi=500,bbox_inches='tight')

    #plot independent distribution of nA and U
    plt.figure(figsize=[3,3])
    for j in range(len(muPAs)):
        nAj = nAs[j]
        cnAj = np.zeros(len(nAbins))
        for nA in nAj:
            idx = (np.abs(nAbins - nA)).argmin()
            nA = float(nAbins[idx])
            id = np.where(abs(nAbins-nA)<1)[0][0]
            cnAj[id] += 1
        plt.plot(nAbins,cnAj,marker=None,ls='-',lw=1,mfc="None",ms=2,label='$\mu_{PA}$'+'={}'.format(muPAs[j]))
    plt.ylim(0)
    plt.xlabel('$\/n_{A-}$')
    plt.ylabel('Count')
    plt.title('k {}'.format(k),loc='center')
    plt.savefig('k{}_nAcount.png'.format(k),dpi=500,bbox_inches='tight')
    plt.legend(loc='best')
    
    plt.figure(figsize=[3,3])
    for j in range(len(muPAs)):
        Uj = Us[j]
        cUj = np.zeros(len(Ubins))
        for U in Uj:
            idx = (np.abs(Ubins - U)).argmin()
            U = float(Ubins[idx])
            id = np.where(abs(Ubins-U)<1)[0][0]
            cUj[id] += 1
        plt.plot(Ubins,cUj,marker=None,ls='-',lw=1,mfc="None",ms=2,label='$\mu_{PA}$'+'={}'.format(muPAs[j]))
    plt.ylim(0)
    plt.xlabel('Potential Energy')
    plt.ylabel('Count')
    plt.title('k {}'.format(k),loc='center')
    plt.savefig('k{}_PEcount.png'.format(k),dpi=500,bbox_inches='tight')
    plt.legend(loc='best')
    plt.show()
