import mdtraj as md
from simtk import openmm, unit
from simtk.unit import *
from simtk.openmm import app
import numpy as np
import simtk.openmm.app.topology as topology

def MC(simulation, state, system, charges, muPA, muPB, temperature, kb, sweeps, steps, stride, pHMC, Polybase, Polyacid):
    deprot = -1
    neutral = 0
    prot = +1
    r_muPA = muPA/(kb*temperature)
    r_muPB = muPB/(kb*temperature)

    fext = 'mu'+str(int(r_muPA))+str(int(r_muPB))+'_'

    report = open(fext+'PEMC.txt','w')
    report.write('# sweep PE\n')
    report.close()

    report = open(fext+'ionization.txt','w')
    report.write('# sweep fA fB\n')
    report.close()

    report = open(fext+'chargesMC.txt','w')
    report.write('# sweep charges\n')
    report.close()

    report = open(fext+'chargesMD.txt','w')
    report.write('# sweep charges\n')
    report.close()

    report = open(fext+'trajectoryMC.pdb','w') #record position every stride MC sweeps
    report.close()

    report = open(fext+'trajectoryMD.pdb','w') #record position after every HMC move
    report.close()

    report = open(fext+'Pacc.txt','w')
    report.write('# sweep Pacc\n')
    report.close()

    Nacc = 0
    traj = fext+'trajectoryMD.dcd'
    dcdReporter = md.reporters.DCDReporter(traj,100)
    dataReporter = app.statedatareporter.StateDataReporter(fext+'log.txt', 500, totalSteps=steps,
                                                       step=True, speed=True, progress=True, potentialEnergy=True, kineticEnergy=True,
                                                       totalEnergy=True, temperature=True, volume=True, density=True, remainingTime=True,separator='\t')
    #==========
    # Simulate
    #==========
    context = simulation.context
    context.setState(state)
    simulation.currentStep = 0
    simulation.reporters.append(dcdReporter)
    simulation.reporters.append(dataReporter)
    for sweep in range(sweeps):
        rand = np.random.rand()
        state1 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
        PE1 = state1.getPotentialEnergy()
        if rand < pHMC:
            simulation.context.setVelocitiesToTemperature(temperature)
            simulation.step(steps)
            state2 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
            PE2 = state2.getPotentialEnergy()
            Pacc = min(1,np.exp(-1/(kb*temperature)*(PE2-PE1)))
            if np.random.rand() > Pacc: #reject and revert to state 1
                context.setState(state1)
            else:
                Nacc += 1
            currentstate = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
            app.PDBFile.writeFile(simulation.topology, currentstate.getPositions(), open(fext+'trajectoryMD.pdb','a'))
            report = open(fext+'chargesMD.txt','a')
            report.write('{} {}\n'.format(sweep,charges))
        else:
            Coulomb = system.getForce(2)
            mon = np.random.randint(Coulomb.getNumParticles())
            charge1 = Coulomb.getParticleParameters(mon)[0]
            if mon in Polyacid:
                if charge1 == deprot:
                    charge2 = neutral
                    Coulomb.setParticleParameters(mon,[charge2])
                    Coulomb.updateParametersInContext(context)
                    state2 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
                    PE2 = state2.getPotentialEnergy()
                    deltaF = PE2 - PE1 + muPA
                elif charge1 == neutral:
                    charge2 = deprot
                    Coulomb.setParticleParameters(mon,[charge2])
                    Coulomb.updateParametersInContext(context)
                    state2 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
                    PE2 = state2.getPotentialEnergy()
                    deltaF = PE2 - PE1 - muPA
            elif mon in Polybase:
                if charge1 == prot:
                    charge2 = neutral
                    Coulomb.setParticleParameters(mon,[charge2])
                    Coulomb.updateParametersInContext(context)
                    state2 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
                    PE2 = state2.getPotentialEnergy()
                    deltaF = PE2 - PE1 + muPB
                elif charge1 == neutral:
                    charge2 = prot
                    Coulomb.setParticleParameters(mon,[charge2])
                    Coulomb.updateParametersInContext(context)
                    state2 = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
                    PE2 = state2.getPotentialEnergy()
                    deltaF = PE2 - PE1 - muPB
            Pacc = min(1,np.exp(-1/(kb*temperature)*deltaF))
            if np.random.rand() > Pacc: #reject and revert to state 1
                context.setState(state1)
                Coulomb.setParticleParameters(mon,[charge1])
                Coulomb.updateParametersInContext(context)
            else:
                Nacc += 1
                charges[mon] = charge2
            currentstate = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)
        if sweep%stride == 0:
            qAs = charges[0:len(Polyacid)]
            qBs = charges[len(Polyacid):]
            try:
                fA = (len(np.where(np.array(qAs,dtype=int) != 0)[0])/len(qAs))
            except:
                fA = 'Nan'
            try:
                fB = (len(np.where(np.array(qBs,dtype=int) != 0)[0])/len(qBs))
            except:
                fB = 'Nan'
            report = open(fext+'PEMC.txt','a')            
            report.write('%i %5.5f\n'%(sweep,currentstate.getPotentialEnergy()/(kb*temperature)))
            report = open(fext+'ionization.txt','a')
            report.write('{} {} {}\n'.format(sweep,fA,fB))
            report = open(fext+'chargesMC.txt','a')
            report.write('{} {} \n'.format(sweep,charges))
            report = open(fext+'Pacc.txt','a')
            report.write('%i %2.2f\n'%(sweep,Nacc/(sweep+1)))
            app.PDBFile.writeFile(simulation.topology, currentstate.getPositions(), open(fext+'trajectoryMC.pdb','a'))
    statef = fext+'output.xml'
    checkf = fext+'checkpoint.chk'
    simulation.saveState(statef)
    simulation.saveCheckpoint(checkf)
    finalstate = simulation.context.getState(getPositions=True, getVelocities=True, getForces=True, getEnergy=True, getParameters=True, getParameterDerivatives=True, enforcePeriodicBox=True)

    #remove data and pdb reporters
    simulation.reporters = simulation.reporters[:-2]
    t = md.load(traj,top=fext+'trajectoryMD.pdb')
    trajOut = traj.split('.')[0] + '.xyz'
    t.save(trajOut)

    t = md.load(fext+'trajectoryMD.pdb')
    t.save(fext+'trajectoryMD.xyz')
    t = md.load(fext+'trajectoryMC.pdb')
    t.save(fext+'trajectoryMC.xyz')

    return statef,checkf,finalstate,charges,simulation
